import os
import json
from .common import MissingFileError

config_file_name = 'config.json'
if 'JUMPER_STAGING' in os.environ:
    config_file_name = 'config.staging.json'
if 'JUMPER_STAGING_INBAR' in os.environ:
    config_file_name = 'config.inbar.json'
JUMPER_DIR = os.path.join(os.path.expanduser('~'), '.jumper')
DEFAULT_CONFIG = os.path.join(JUMPER_DIR, config_file_name)


def load_config(config_file=None):
    config_file = config_file or DEFAULT_CONFIG
    if not os.path.isfile(config_file):
        raise MissingFileError("File not found: {}".format(config_file))
    with open(config_file) as config_data:
        return json.load(config_data)


def get_token(config_file=None):
    config = load_config(config_file)
    if 'token' in config:
        return config['token']
    else:
        raise ValueError("token is missing in: {}".format(config_file))


def set_token(token, config_file=None):
    config_file = config_file or DEFAULT_CONFIG
    config = dict()
    try:
        config = load_config(config_file)
    except MissingFileError:
        pass

    config['token'] = token
    with open(config_file, 'w') as file:
        file.write(json.dumps(config))



"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
import os
import os.path
import array
from time import sleep
from threading import Thread, Lock
import socket
import errno
import socket
import json
import struct
import logging
import sys

from .timeout_dec import timeout
from .common import TimeoutException
from common import WrongStateError



class JemuUartTcp(object):
    """
    Represents a UART device.
    """
    # uart_port = 0
    def __init__(self, vlab):
        self._buffer = b''
        self._vlab = vlab
        # self._read_from_uart_thread = None
        self._lock = Lock()
        # self._print_to_screen = False
        # self._should_close_thread = False

        self._addr = "localhost"
        self._conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._conn.settimeout(0.3)
        self._callbacks = []
        self._thread = None
        self._should_run = False
        self._connection_closed = True
        self._threads = []

    # def print_uart_to_screen(self):
    #     self._print_to_screen = True

    def open(self):
        self._should_run = True
        t = Thread(target=self._connection_task)
        self._threads.append(t)
        t.start()

        @timeout(6)
        def wait_for_connection():
            while self._uart_port == 0:
                sleep(0.1)

            while not self._connect():
                sleep(0.1)
        try:
            wait_for_connection()
        except TimeoutException:
            self._vlab.stop()
            raise Exception("Error: Couldn't connect to Emulator. Please contact us at support@jumper.io with a copy of this trace text port: " + self._uart_port)

    def remove(self):
        pass

    def _connect(self):
        result = False

        try:
            self._conn.connect((self._addr, int(self._uart_port)))
            self._connection_closed = False
            result = True
        except Exception:
            self._conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        finally:
            return result

    def close(self):
        self._should_run = False
        for t in self._threads:
            t.join()

    def _connection_task(self):
        while self._should_run:
            try:
                new_data = self._conn.recv(1024)
                if new_data == '':
                    break

                # sys.stdout.write(new_data)
                
                with self._lock:
                    self._buffer += new_data
            except socket.timeout:
                continue
            except socket.error as e:
                self._connection_closed = True
                return None

        if not self._connection_closed:
            self._conn.close()
            self._connection_closed = True
   
    def read_line(self, line_separator=b'\r\n'):
        """
        Reads a single line from UART.

        :param line_separator: Substring for separating lines. Usually b'\\\\r\\\\n' or b'\\\\n'

        :return: A byte string of data read from the device, including line_separator
        """
        while line_separator not in self._buffer:
            sleep(0.2)
        with self._lock:
            line_length = self._buffer.find(line_separator) + len(line_separator)
            data = self._buffer[:line_length]
            self._buffer = self._buffer[line_length:]
        return data

    def read(self):
        """

        :return: data available on the UART device. Returns an empty bytes string if no data is available.
        """
        with self._lock:
            data = self._buffer
            self._buffer = b''
        return data

    def wait_until_uart_receives(self, data, timeout=None):
        """
        Blocks until specified data is received on UART.
        If the device is paused, this function will continue the execution and will pause the device when the data is ready or timeout occured.
        If the device is running, the function will not pause the device.

        :param data: Expected bytes string.
        :param timeout: Emulation time specified in milliseconds. jumper.vlab.TimeoutException is raised if timeout occurred before data was available in the buffer.
        :return: The data available on in the UART buffer when the specified data was received. Note that this can be more than the data provided in the data parameter.
        """
        original_state = self._vlab.get_state()

        if original_state == 'running':
            self._vlab.pause()
        elif original_state != 'paused':
            raise WrongStateError('state should be paused or running but was {}'.format(original_state))

        if timeout is not None:
            self._vlab.stop_after_ms(timeout)

        self._vlab.resume()

        while data not in self._buffer and self._vlab.get_state() == 'running':
            sleep(0.2)

        if timeout is not None:
            self._vlab.cancel_stop()

        if original_state == 'running':
            self._vlab.resume()
        elif original_state == 'paused':
            self._vlab.pause()
        with self._lock:
            if data in self._buffer:
                tmp = self._buffer
                self._buffer = b''
                return tmp

        raise TimeoutException("timeout while waiting for data from uart")

    def write(self, data):
        self._conn.sendall(data)



import re
import sys
from time import sleep
from .common import VlabException
from .vlab import Vlab
from threading import Event


class VlabRunner:
    def __init__(self):
        self.stop = Event()
        self.stop.clear()
        pass

    def start(self, args):
        try:
            reached_bkpt = Event()

            def bkpt_callback(code):
                print('Firmware reached a BKPT instruction with code {}'.format(code))
                reached_bkpt.set()

            vlab = VlabRunner.init_vlab(args)
            vlab.on_bkpt(bkpt_callback)
            ns = None

            if args.stop_after:
                vlab._send_stop_after_event()

                def raise_invalid_time():
                    print('Invalid time format for --stop-after: {}'.format(args.stop_after))
                    return 1

                match = re.match('^\d+', args.stop_after)
                if not match:
                    raise_invalid_time()
                ns = int(match.group(0))

                if re.match('^\d+$', args.stop_after) or re.match('^\d+ms$', args.stop_after):
                    ns = ns * 1000 * 1000
                elif re.match('^\d+s$', args.stop_after):
                    ns = ns * 1000 * 1000 * 1000
                elif re.match('^\d+us$', args.stop_after):
                    ns = ns * 1000
                elif re.match('^\d+ns$', args.stop_after):
                    pass
                else:
                    raise_invalid_time()

            if not args.print_uart and not args.traces_list:
                print(
                    '\nVirtual device is running without UART/Trace prints (use -u and/or -t to get your firmware '
                    'execution status)\n')
            else:
                print('\nVirtual device is running\n')
                sys.stdout.flush()

            vlab.start(ns)

            try:
                while (not reached_bkpt.is_set()) and not self.stop.is_set() and vlab.get_state() == 'running':
                    sleep(0.5)
            except KeyboardInterrupt:
                pass

            vlab.stop()
            return vlab.get_return_code()

        except VlabException as e:
            print('jumper run: error: {}'.format(e.message))
            try:
                if vlab:
                    vlab.stop()
            finally:
                return e.exit_code

    def stop_wait(self):
        while not self.stop.wait(timeout=0.1):
            pass

    def stop_run(self):
        self.stop.set()

    @staticmethod
    def init_vlab(args):
        registers_trace = False
        functions_trace = False
        interrupts_trace = False
        if args.traces_list:
            for trace in args.traces_list.split(','):
                if trace == 'registers' or trace == 'regs':
                    registers_trace = True
                elif trace == 'functions':
                    functions_trace = True
                elif trace == 'interrupts':
                    interrupts_trace = True
                else:
                    raise VlabException(
                        'jumper run: error: Invalid trace type {}. Valid traces are regs, functions, interrupts'.format(
                            trace), 1)

        vlab = Vlab(
            working_directory=args.working_directory,
            sudo_mode=args.sudo_mode,
            gdb_mode=args.gdb_mode,
            registers_trace=registers_trace,
            functions_trace=functions_trace,
            interrupts_trace=interrupts_trace,
            trace_output_file=args.trace_output_file,
            print_uart=args.print_uart,
            uart_output_file=args.uart_output_file,
            uart_device=args.uart_device,
            platform=args.platform,
            token=args.token,
            debug_peripheral=args.debug_peripheral,
            gpio=args.gpio
        )

        vlab.load(args.fw)

        return vlab







"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from threading import Thread, Lock
import threading
import socket
import json
import struct
import logging
import sys
import queue
from time import sleep
from common import JemuConnectionException


class JemuConnection:
    _addr = None
    _conn = None
    _HANDSHAKE = "handshake"
    _COMMAND_TYPE = "command_type"
    _TRANSITION_TYPE = "transition_type"
    _PIN_NUM = "pin_number"
    _MESSAGE_ID = "message_id"
    _VOLTAGE = "voltage"
    _API_VERSION = "1"
    _OK = "ok"
    _ERROR = "error"
    _ERROR_CODE = "404"
    _TYPE = "type"
    _RESPONSE = "response"
    _COMMAND = "command"
    _EVENT = "event"
    _API_V_STRING = "api_version"
    _ERROR_MESSAGE = "error"
    _COMMAND_START = "start"
    _NANO_SECONDS = "nanoseconds"
    _COMMAND_SET = "command_set"
    _COMMAND_GET = "command_get"
    _PERIPHERAL_ID = "peripheral_id"
    _PERIPHERAL_TYPE = "peripheral_type"

    _LOG_LEVEL = logging.ERROR
    _logger = logging.getLogger('JemuConnection')
    log_handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    log_handler.setFormatter(formatter)
    _logger.addHandler(log_handler)
    _logger.setLevel(_LOG_LEVEL)
    log_handler.setLevel(_LOG_LEVEL)

    def __init__(self, addr):
        self._addr = addr
        self._conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._conn.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._conn.settimeout(0.3)
        self._callbacks = []
        self._should_stop_running = threading.Event()
        self._connection_closed = True
        self._response_packet_received = threading.Event()
        self._response_packet = None
        self._threads = []
        self.events_queue = queue.Queue()
        self._lock = Lock()

    def start(self):
        self._should_stop_running.clear()
        t = Thread(target=self.connection_task)
        self._threads.append(t)
        t.start()
        t_events = Thread(target=self.handle_events)
        self._threads.append(t_events)
        t_events.start()

    def close(self):
        self._should_stop_running.set()
        for t in self._threads:
            t.join()

    def is_connected(self):
        return not self._should_stop_running.is_set()

    def connection_task(self):
        while not self._should_stop_running.is_set():
            json_buff_string = self.recv_json()
            if json_buff_string is None:
                self._should_stop_running.set()
                break
            self._logger.debug("received data: " + json_buff_string)
            try:
                json_pack = json.loads(json_buff_string)
            except ValueError as e:
                print('Error parsing json buffer: {}'.format(json_buff_string))
                raise e

            if self._TYPE in json_pack:
                if json_pack[self._TYPE] == self._EVENT:
                    self.events_queue.put(json_pack)
                elif json_pack[self._TYPE] == self._RESPONSE:
                    self._response_packet = json_pack
                    self._response_packet_received.set()
            else:
                self._should_stop_running.set()
                raise JemuConnectionException("Error missing 'type' key in message.")

        if not self._connection_closed:
            self._conn.close()
            self._connection_closed = True

    def inform(self, jemu_json_packet):
        for callback in self._callbacks:
            callback(jemu_json_packet)

    def handle_events(self):
        while not (self._should_stop_running.is_set() and self.events_queue.empty()):
            try:
                event = self.events_queue.get(timeout=0.5)
            except queue.Empty:
                continue
            self.inform(event)
            self.events_queue.task_done()

    def register(self, callback):
        self._callbacks.append(callback)

    def connect(self, port):
        result = False
        try:
            self._conn.connect((self._addr, int(port)))
            self._connection_closed = False
            result = True
        except Exception:
            self._conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._conn.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        finally:
            return result

    def handshake(self, ns=None):
        self._should_stop_running.clear()
        self.send_handshake()
        response_ack = self.recv_json()
        if response_ack is None:
            return False
        self.check_response(response_ack)
        self.send_start(ns)
        return True

    def check_response(self, response_ack):
        self._logger.debug("handshake response data: " + response_ack)
        response = json.loads(response_ack)
        message_type = response[self._TYPE]
        if message_type != self._RESPONSE:
            self._should_stop_running.set()
            raise JemuConnectionException("Error expect response message instead got [" + message_type + "]")
        status_code = response[self._RESPONSE]
        if status_code != self._OK:
            self._should_stop_running.set()
            raise JemuConnectionException("Sdk's api version does not match to the jemu's version")

    def send_json(self, json_obj):
        data_to_send = json.dumps(json_obj)
        number_of_bytes = len(data_to_send.encode('utf-8'))
        self._logger.debug("Number of byte of send data: " + str(number_of_bytes))
        number_to_send = struct.pack('!i', number_of_bytes)
        self.send(number_to_send)
        self._logger.debug("Data to send: " + data_to_send)
        self.send(data_to_send)

    def send_handshake(self):
        hand_shake_message = {self._API_V_STRING: self._API_VERSION, self._TYPE: self._HANDSHAKE}
        self.send_json(hand_shake_message)

    def send_start(self, ns=None):
        ack = {self._TYPE: self._COMMAND, self._COMMAND: self._COMMAND_START}
        if ns:
            ack[self._NANO_SECONDS] = ns
        self.send_json(ack)

    def send_command(self, json_pack, peripheral_id, peripheral_type):
        json_pack[self._TYPE] = self._COMMAND
        json_pack[self._PERIPHERAL_ID] = peripheral_id
        json_pack[self._PERIPHERAL_TYPE] = peripheral_type

        with self._lock:
            self._response_packet_received.clear()
            self.send_json(json_pack)
            while not (self._response_packet_received.is_set() or self._should_stop_running.is_set()):
                self._response_packet_received.wait(0.5)

            if self._should_stop_running.is_set():
                return None
            return self._response_packet

    def send_command_async(self, json_pack, peripheral_id, peripheral_type):
        json_pack[self._TYPE] = self._COMMAND
        json_pack[self._PERIPHERAL_ID] = peripheral_id
        json_pack[self._PERIPHERAL_TYPE] = peripheral_type
        self.send_json(json_pack)

    def send(self, buffer):
        try:
            self._conn.sendall(buffer)
        except socket.error as e:
            self._should_stop_running.set()
            raise JemuConnectionException("Jemu connection failed with error: " + str(e))

    def clean_threat_list(self):
        for t in self._threads:
            if not t.isAlive():
                self._threads.remove(t)

    def recv_json(self):
        data = self.receive(4)
        if data is None:
            return None
        buffer_size = struct.unpack("!i", data)[0]
        self._logger.debug("Number of byte of received data: " + str(buffer_size))
        if buffer_size == 0:
            return None
        data = self.receive(buffer_size)
        if data is None:
            return None
        else:
            return data

    def receive(self, size):
        result = b''
        while (len(result) < size) and (not self._connection_closed):
            if self._should_stop_running.is_set():
                return None

            try:
                data = self._conn.recv(size - len(result))
            except socket.timeout:
                continue
            except socket.error as e:
                sleep(0.2)
                if not self._should_stop_running.is_set():
                    self._should_stop_running.set()
                    raise JemuConnectionException("Jemu connection failed with error: " + str(e))
                else:
                    self._connection_closed = True
                    return None

            if data == b'':
                self._connection_closed = True
                return None

            result += data

        return result


"""
:copyright: (c) 2017 by Jumper Labs Ltd.
:license: Apache 2.0, see LICENSE.txt for more details.
"""
from __future__ import print_function

import signal
import sys

from .vlab import Vlab
from __version__ import __version__ as jumper_current_version
from .common import VlabException
from .jemu_args_parser import JemuArgsParser
from .vlab_hci_device import VirtualHciDevice
from .vlab_runner import VlabRunner
from .login import login

vlab_runner = VlabRunner()


def signal_handler(signum, frame):
    vlab_runner.stop_run()

signal.signal(signal.SIGTERM, signal_handler)
signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGABRT, signal_handler)


def validate_traces_list(traces_list):
    traces_list = traces_list.split(',')
    for trace in traces_list:
        if trace not in ['regs', 'interrupts', 'functions']:
            raise VlabException('jumper run: error: unrecognized trace:' + trace, 1)


def run(args):
    vlab_runner.start(args)


def hci():
    Vlab.check_sdk_version(False)
    virtual_hci_device = VirtualHciDevice()
    
    virtual_hci_device.start()
    vlab_runner.stop_wait()

    virtual_hci_device.stop()


def command_run(args):
    if args.version:
        print("v" + jumper_current_version)
        return 0

    if not args.fw:
        print("jumper run: error: argument --bin/-b/--fw is required")
        return 1

    if args.traces_list:
        args.traces_list = args.traces_list.lower().replace(' ', '')
        validate_traces_list(args.traces_list)

    return run(args)


def main():
    jemu_parser = JemuArgsParser()
    jemu_parser.parse(sys.argv[1:])
    args = jemu_parser.get_args()
    if args.command == 'run':
        return command_run(args)

    elif args.command == 'ble':
        hci()
        return 0

    elif args.command == 'login':
        login()


if __name__ == '__main__':
    try:
        exit(main())
    except VlabException as e:
        print(e.message)
        exit(e.exit_code())



import tornado.ioloop
import tornado.web
import json
import webbrowser
import socket

from .config import set_token, DEFAULT_CONFIG

token = None


class TokenHandler(tornado.web.RequestHandler):
    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "*")
        self.set_header('Access-Control-Allow-Methods', 'POST, OPTIONS')

    def post(self):
        global token
        data = json.loads(self.request.body)
        token = data['token']
        self.write('OK')
        tornado.ioloop.IOLoop.current().stop()

    def options(self):
        self.set_status(204)
        self.finish()


def token_server():
    global token
    app = tornado.web.Application([(r"/token", TokenHandler)])
    port = choose_port()
    server = app.listen(port)
    webbrowser.open_new('https://vlab.jumper.io/cli_login/{}'.format(port))
    tornado.ioloop.IOLoop.current().start()
    server.stop()
    return token


def choose_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    addr, port = s.getsockname()
    s.close()
    return port


def login():
    token = token_server()
    set_token(token)
    print("Token has been saved to {}".format(DEFAULT_CONFIG))
    print("Login successful")
    return token


if __name__ == "__main__":
    print(token_server())
